"""TotalConnectClient() in this file is the primary class of this package.

Instantiate it like this:

usercodes = { 'default': '1234' }
client = TCC.TotalConnectClient(username, password, usercodes)

for location in client.locations:
    ### do stuff with this location
"""

import base64
import logging
import ssl
from ssl import SSLContext
import time
from importlib import resources as impresources
from typing import Dict, Any
import requests
import urllib3.poolmanager
import zeep
import zeep.cache
import zeep.transports
from zeep.exceptions import Fault as ZeepFault
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5
import jwt

from . import cache as cache_folder
from .const import ArmType, _ResultCode
from .exceptions import (
    AuthenticationError,
    BadResultCodeError,
    FailedToBypassZone,
    FeatureNotSupportedError,
    InvalidSessionError,
    RetryableTotalConnectError,
    ServiceUnavailable,
    TotalConnectError,
    UsercodeInvalid,
    UsercodeUnavailable,
)
from .location import TotalConnectLocation
from .user import TotalConnectUser

DEFAULT_USERCODE = "-1"

SCHEMAS_TO_CACHE = {
    "https://schemas.xmlsoap.org/soap/encoding/": "soap-encodings-schemas.xmlsoap.org.txt",
}

HTTP_REQUEST_TIMEOUT = 30

LOGGER = logging.getLogger(__name__)


class _SslContextAdapter(requests.adapters.HTTPAdapter):
    """Makes Zeep use our ssl_context."""

    def __init__(self, ssl_context: SSLContext, **kwargs) -> None:
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(
        self, num_pools: int, maxsize: int, block: bool = False
    ) -> None:
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=num_pools,
            maxsize=maxsize,
            block=block,
            ssl_context=self.ssl_context,
        )


class TotalConnectClient:
    """Client for Total Connect."""

    TIMEOUT = 60  # seconds until SOAP I/O will fail

    def __init__(  # pylint: disable=too-many-arguments
        self,
        username: str,
        password: str,
        usercodes: Dict[str, str] | None = None,
        auto_bypass_battery: bool = False,
        retry_delay: int = 6,  # seconds between retries
    ) -> None:
        """Initialize."""
        self.times = {}
        self.time_start = time.time()
        self.soap_client = None

        self.username: str = username
        self.password: str = password
        self.usercodes = usercodes or {}
        self.auto_bypass_low_battery = auto_bypass_battery
        self.retry_delay = retry_delay

        self.token = None
        self._client_id = None
        self._session_expiration = None
        self._refresh_token = None
        self._invalid_credentials = False

        self._module_flags: Dict[str, str] = {}
        self._user: TotalConnectUser | None = None
        self._locations: Dict[Any, TotalConnectLocation] = {}
        self._locations_unfetched: Dict[Any, TotalConnectLocation] = {}

        self.authenticate()

        self.times["__init__"] = time.time() - self.time_start

    @property
    def locations(self) -> Dict[Any, TotalConnectLocation]:
        """Raises an exception if the panel cannot be reached to retrieve
        metadata or details. This can be retried later and will succeed
        if/when the panel becomes reachable.
        """
        # to_fetch is needed because items() is invalidated by del
        to_fetch = list(self._locations_unfetched.items())
        for locationid, location in to_fetch:
            try:
                location.get_partition_details()
                location.get_zone_details()
                location.get_panel_meta_data()
                # if we get here, it has been fetched successfully
                del self._locations_unfetched[locationid]
            except Exception:
                LOGGER.error(f"exception during initial fetch of {locationid}")
                raise
        assert not self._locations_unfetched
        return self._locations

    def __str__(self) -> str:
        """Return a text string that is printable."""
        data = (
            f"CLIENT\n\n"
            f"Username: {self.username}\n"
            f"Password: {'[hidden]' if self.password else '[unset]'}\n"
            f"Usercode: {self.usercodes}\n"
            f"Auto Bypass Low Battery: {self.auto_bypass_low_battery}\n"
            f"Invalid Credentials: {self._invalid_credentials}\n"
            f"Module Flags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + str(self._user)

        locations = f"LOCATIONS: {len(self.locations)}\n\n"
        for location in self.locations:
            locations = locations + str(self.locations[location])

        return data + locations

    def times_as_string(self) -> str:
        """Return a string with times."""
        self.times["total running time"] = time.time() - self.time_start
        msg = "total-connect-client time info (seconds):\n"
        for key, value in self.times.items():
            msg = msg + f"  {key}: {value}\n"

        return msg

    def _raise_for_retry(self, response: Dict[str, Any]) -> None:
        """Determine which responses should be retried in request()."""
        rc = _ResultCode.from_response(response)
        if rc == _ResultCode.INVALID_SESSION:
            raise InvalidSessionError("invalid session", response)
        if rc == _ResultCode.INVALID_SESSIONID:
            raise InvalidSessionError("invalid session ID", response)
        if rc == _ResultCode.CONNECTION_ERROR:
            raise RetryableTotalConnectError("connection error", response)
        if rc == _ResultCode.FAILED_TO_CONNECT:
            raise RetryableTotalConnectError("failed to connect with panel", response)
        if rc == _ResultCode.BAD_OBJECT_REFERENCE:
            raise RetryableTotalConnectError("bad object reference", response)

    def raise_for_resultcode(self, response: Dict[str, Any]) -> None:
        """If response.ResultCode indicates success, return and do nothing.

        If it indicates an authentication error, raise AuthenticationError.
        """
        rc = _ResultCode.from_response(response)
        if rc in (
            _ResultCode.SUCCESS,
            _ResultCode.ARM_SUCCESS,
            _ResultCode.DISARM_SUCCESS,
            _ResultCode.SESSION_INITIATED,
        ):
            return
        self._raise_for_retry(response)
        if rc in (
            _ResultCode.BAD_USER_OR_PASSWORD,
            _ResultCode.AUTHENTICATION_FAILED,
            _ResultCode.ACCOUNT_LOCKED,
        ):
            raise AuthenticationError(rc.name, response)
        if rc == _ResultCode.USER_CODE_UNAVAILABLE:
            raise UsercodeUnavailable(rc.name, response)
        if rc == _ResultCode.USER_CODE_INVALID:
            raise UsercodeInvalid(rc.name, response)
        if rc == _ResultCode.FEATURE_NOT_SUPPORTED:
            raise FeatureNotSupportedError(rc.name, response)
        if rc == _ResultCode.FAILED_TO_BYPASS_ZONE:
            raise FailedToBypassZone(rc.name, response)
        raise BadResultCodeError(rc.name, response)

    def _send_one_request(
        self, operation_name: str, args: list[Any] | tuple[Any, ...]
    ) -> Dict[str, Any]:
        LOGGER.debug(f"sending API request {operation_name}{args}")
        operation_proxy = self.soap_client.service[operation_name]
        return zeep.helpers.serialize_object(operation_proxy(*args))

    API_ENDPOINT = "https://rs.alarmnet.com/TC21api/tc2.asmx?WSDL"
    CONFIG_ENDPOINT = "https://totalconnect2.com/application.config.json"
    TOKEN_ENDPOINT = "https://rs.alarmnet.com/TC2API.Auth/token"

    def request(
        self,
        operation_name: str,
        args: list[Any] | tuple[Any, ...],
        attempts_remaining: int = 5,
    ) -> Dict[str, Any]:
        """Send a SOAP request.

        args is a list or tuple defining the parameters to the operation.
        """
        is_first_request = attempts_remaining == 5
        attempts_remaining -= 1
        if not self.soap_client:
            # the server doesn't support RFC 5746 secure renegotiation, which
            # causes OpenSSL to fail by default. here we override the default.
            # if they fix their server, we should revert this change to ctx.options
            session = requests.Session()
            ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            ctx.options |= 0x04  # ssl.OP_LEGACY_SERVER_CONNECT once that exists
            session.mount("https://", _SslContextAdapter(ctx))
            cache = zeep.cache.InMemoryCache()
            for url, filename in SCHEMAS_TO_CACHE.items():
                cache_file = impresources.files(cache_folder) / filename
                with cache_file.open() as file:
                    cache.add(url, file.read())
            transport = zeep.transports.Transport(
                session=session,
                cache=cache,
                timeout=self.TIMEOUT,  # for loading WSDL and xsd documents
                operation_timeout=self.TIMEOUT,  # for operations (POST/GET)
            )
            self.soap_client = zeep.Client(self.API_ENDPOINT, transport=transport)
        try:
            # Refresh session if needed. If the refresh fails it will throw an InvalidSessionError,
            # forcing a reauthentication
            self._check_and_refresh_session()

            response = self._send_one_request(operation_name, args)
            self._raise_for_retry(response)
            return response
        # To retry an exception that could be raised during the
        # request, add it to an except block here, depending on what
        # you want to have happen. The first block just retries and
        # logs. The second block causes reauthentication.
        except (
            RetryableTotalConnectError,
            requests.exceptions.RequestException,
        ) as err:
            if attempts_remaining <= 0:
                raise
            if isinstance(err, RetryableTotalConnectError):
                msg = (
                    f"{self.username} {operation_name}{args} {err.args[0]} on response"
                )
            else:
                msg = f"{self.username} {operation_name}{args} {err} on request"
            if is_first_request:
                LOGGER.info(f"{msg}: {attempts_remaining} retries remaining")
            else:
                LOGGER.debug(f"{msg}: {attempts_remaining} retries remaining")
            time.sleep(self.retry_delay)
        except (ZeepFault, requests.exceptions.HTTPError) as err:
            if attempts_remaining <= 0:
                raise ServiceUnavailable(
                    f"Error connecting to Total Connect service: {err}"
                ) from err
            LOGGER.debug(
                f"Error connecting to Total Connect service: {attempts_remaining} retries remaining"
            )
            time.sleep(self.retry_delay)
        except InvalidSessionError as err:
            if attempts_remaining <= 0:
                raise ServiceUnavailable(
                    f"Invalid Session after multiple retries: {err}"
                ) from err
            LOGGER.info(
                f"reauthenticating {self.username}: {attempts_remaining} retries remaining"
            )
            old_token = self.token
            self.token = None
            self.authenticate()
            args = [self.token if old_token == arg else arg for arg in args]
        return self.request(operation_name, args, attempts_remaining)

    def _encrypt_credential(self, credential: str, key_pem: str) -> str:
        # Load the key from the PEM file
        key = RSA.importKey(key_pem)

        # Create a cipher object using PKCS1_OAEP padding
        cipher = PKCS1_v1_5.new(key)

        # Encrypt the message
        encrypted_message = cipher.encrypt(credential.encode())

        # Encode the encrypted message in base64 for safe transmission
        return base64.b64encode(encrypted_message).decode()

    def authenticate(self) -> None:
        """Login to the system.

        Upon success, self.token is a valid credential
        for further API calls, and self._user and self.locations are valid.
        self.locations will not be refreshed if it was non-empty on entry.
        """
        start_time = time.time()
        if self._invalid_credentials:
            raise AuthenticationError(
                f"not authenticating: password already failed for user {self.username}"
            )

        # Retrieve application configuration for TotalConnect web app, this is needed for the
        # encryption key and various IDs.
        response = requests.get(self.CONFIG_ENDPOINT, timeout=HTTP_REQUEST_TIMEOUT)
        if not response.ok:
            raise ServiceUnavailable(
                f"Service configuration is not available at {self.CONFIG_ENDPOINT}"
            )
        config = response.json()
        key = config["AppConfig"][0]["tc2APIKey"]
        self._client_id = config["AppConfig"][0]["tc2ClientId"]
        locale = 'en-US'
        app_id = next(
            info for info in config["brandInfo"] if info["BrandName"] == "totalconnect"
        )["AppID"]
        app_version = config["RevisionNumber"] + "." + config["version"].split(".")[-1]

        # Encrypt username and password and log in to get a JWT with a session ID.
        key_pem = '-----BEGIN PUBLIC KEY-----\n' + key + '\n-----END PUBLIC KEY-----'
        data = {
            'username': self._encrypt_credential(self.username, key_pem),
            'password': self._encrypt_credential(self.password, key_pem),
            'grant_type': 'password',
            'client_id': self._client_id,
            'locale': locale,
        }
        response = requests.post(
            url=self.TOKEN_ENDPOINT,
            data=data,
            timeout=HTTP_REQUEST_TIMEOUT
        )
        response_json = response.json()
        if not response.ok:
            try:
                self.raise_for_resultcode(response_json)
            except AuthenticationError:
                self._invalid_credentials = True
                self.token = None
                raise
        jwt_token = jwt.decode(
            response_json['access_token'],
            algorithms="HS256",
            options={"verify_signature": False}
        )
        self.token = jwt_token["ids"].split(';', 1)[0]

        # We will need to refresh the token periodically, keep track of the expiration
        # time using the system's monotonic clock so that we aren't affected by clock
        # updates.
        self._refresh_token = response_json["refresh_token"]
        self._session_expiration = time.monotonic() + int(response_json["expires_in"])

        # Retrieve user and location information.
        if not self._locations:
            response = self.request(
                "GetSessionDetails",
                (self.token, app_id, app_version)
            )
            self._module_flags = dict(
                x.split("=") for x in response["ModuleFlags"].split(",")
            )
            self._user = TotalConnectUser(response["UserInfo"])
            self._locations_unfetched = self._make_locations(response)
            self._locations = self._locations_unfetched.copy()
            if not self._locations:
                raise TotalConnectError("no locations found", response)
        LOGGER.info(f"{self.username} authenticated: {len(self._locations)} locations")
        self.times["authenticate()"] = time.time() - start_time

    def _check_and_refresh_session(self) -> None:
        """Refresh the current session if needed."""
        if not self._refresh_token:
            raise InvalidSessionError("Not logged in")

        # Check the current time against the session expire time. Use a one second
        # grace period to mitigate race conditions.
        if time.monotonic() < self._session_expiration - 1:
            return

        # Send a refresh request and store the new refresh token and expiration time
        LOGGER.debug("Session has expired, refreshing now")
        data = {
            'refresh_token': self._refresh_token,
            'grant_type': 'refresh_token',
            'client_id': self._client_id,
        }
        response = requests.post(
            url=self.TOKEN_ENDPOINT,
            data=data,
            timeout=HTTP_REQUEST_TIMEOUT
        )
        if not response.ok:
            raise InvalidSessionError("Failed to refresh session")
        response_json = response.json()
        self._refresh_token = response_json["refresh_token"]
        self._session_expiration = time.monotonic() + int(response_json["expires_in"])

    def validate_usercode(self, device_id: str, usercode: str) -> bool:
        """Return True if the usercode is valid for the device."""
        response = self.request("ValidateUserCode", (self.token, device_id, usercode))
        try:
            self.raise_for_resultcode(response)
        except UsercodeInvalid:
            LOGGER.warning(f"usercode {usercode} invalid for device {device_id}")
            return False
        except UsercodeUnavailable:
            LOGGER.warning(f"usercode {usercode} unavailable for device {device_id}")
            return False
        return True

    def is_logged_in(self) -> bool:
        """Return true if the client is logged in to Total Connect."""
        return self.token is not None

    def log_out(self) -> None:
        """Upon return, we are logged out.

        Raises TotalConnectError if we still might be logged in.
        """
        if self.is_logged_in():
            response = self.request("Logout", (self.token,))
            self.raise_for_resultcode(response)
            LOGGER.info("Logout Successful")
            self.token = None

    def get_number_locations(self) -> int:
        """Return the number of locations.

        Home Assistant needs a way to force the locations to load
        inside a callable function.
        """
        return len(self.locations)

    def _make_locations(
        self, response: Dict[str, Any]
    ) -> Dict[Any, TotalConnectLocation]:
        """Return a dict mapping LocationID to TotalConnectLocation."""
        start_time = time.time()
        new_locations = {}

        for locationinfo in (response["Locations"] or {}).get("LocationInfoBasic", {}):
            location_id = locationinfo["LocationID"]
            location = TotalConnectLocation(locationinfo, self)
            new_locations[location_id] = location

            location.auto_bypass_low_battery = self.auto_bypass_low_battery

            # set the usercode for the location
            usercode = (
                self.usercodes.get(location_id)  # noqa: W504
                or self.usercodes.get(str(location_id))
                or self.usercodes.get("default")
            )
            if usercode:
                location.usercode = usercode
            else:
                LOGGER.warning(f"no usercode for location {location_id}")
                location.usercode = DEFAULT_USERCODE

        self.times["_make_locations()"] = time.time() - start_time
        return new_locations


class ArmingHelper:
    """
    For a partition or location, you can call its arm() or disarm() method directly.

    Example: partition.arm(ArmType.AWAY)

    Alternatively, you can use ArmingHelper.
       Example: ArmingHelper(partition).arm_away()
    """

    def __init__(self, partition_or_location) -> None:
        """Initialize ArmingHelper."""
        self.armable = partition_or_location

    def arm_away(self, usercode: str = "") -> None:
        """Arm the system (Away)."""
        self.armable.arm(arm_type=ArmType.AWAY, usercode=usercode)

    def arm_stay(self, usercode: str = "") -> None:
        """Arm the system (Stay)."""
        self.armable.arm(arm_type=ArmType.STAY, usercode=usercode)

    def arm_stay_instant(self, usercode: str = "") -> None:
        """Arm the system (Stay - Instant)."""
        self.armable.arm(arm_type=ArmType.STAY_INSTANT, usercode=usercode)

    def arm_away_instant(self, usercode: str = "") -> None:
        """Arm the system (Away - Instant)."""
        self.armable.arm(arm_type=ArmType.AWAY_INSTANT, usercode=usercode)

    def arm_stay_night(self, usercode: str = "") -> None:
        """Arm the system (Stay - Night)."""
        self.armable.arm(arm_type=ArmType.STAY_NIGHT, usercode=usercode)

    def disarm(self, usercode: str = "") -> None:
        """Disarm the system."""
        self.armable.disarm(usercode=usercode)



"""Test your system from the command line."""

import logging
import sys
from pprint import pprint

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) != 7:
    print(
        "usage:  username password AutomationDeviceID DeviceID DeviceTypeID LocationID\n"
    )
    sys.exit()

USERNAME = sys.argv[1]
PASSWORD = sys.argv[2]
AUTOMATION_DEVICE = sys.argv[3]
DEVICE_ID = sys.argv[4]
TYPE = sys.argv[5]
LOCATION = sys.argv[6]

TC = TotalConnectClient(USERNAME, PASSWORD)

print("\n\n\nGetAutomationDeviceStatus\n\n\n")
try:
    result = TC.request(
        "GetAutomationDeviceStatus", 
        (TC.token, AUTOMATION_DEVICE)
    )
    pprint(result)
except Exception as err:
    print(err)



print("\n\n\nGetAutomationDeviceStatusExV1\n\n\n")
try:
    result = TC.request(
        "GetAutomationDeviceStatusExV1",
        (TC.token, AUTOMATION_DEVICE)
    )
    pprint(result)
except Exception as err:
    print(err)


print("\n\n\nGetAllAutomationDeviceStatusExV1\n\n\n")
try:
    result = TC.request(
        "GetAllAutomationDeviceStatusExV1",
        (TC.token, AUTOMATION_DEVICE)
    )
    pprint(result)
except Exception as err:
    print(err)

print("\n\n\nGetDeviceStatus\n\n\n")

a = []
a.append(DEVICE_ID)
a.append(TYPE)
b = []
b.append(a)

try:
    result = TC.request(
        "GetDeviceStatus",
        (TC.token, LOCATION, b)
    )
    pprint(result)
except Exception as err:
    print(err)

print("\n\n\nGetAutomationLocks\n\n\n")

try:
    result = TC.request(
        "GetAutomationLocks",
        (TC.token, AUTOMATION_DEVICE, DEVICE_ID)
    )
    pprint(result)
except Exception as err:
    print(err)




"""Test total_connect_client."""

import unittest
from unittest.mock import patch

import pytest
from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    RESPONSE_ARMED_AWAY,
    RESPONSE_DISARMED,
    RESPONSE_FEATURE_NOT_SUPPORTED,
)

from total_connect_client.exceptions import (
    FeatureNotSupportedError,
    PartialResponseError,
)

RESPONSE_DISARMED_NONE = {"ResultCode": 0}


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.client = create_client()
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]
        self.location = self.client.locations[self.location_id]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_get_panel_meta_data_normal(self):
        """Test get_panel_meta_data() with a normal response."""
        responses = [RESPONSE_ARMED_AWAY, RESPONSE_DISARMED]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start disarmed
            assert self.location.arming_state.is_disarmed()

            # first response shows armed_away
            self.location.get_panel_meta_data()
            assert self.location.arming_state.is_armed_away()

            # second response shows disarmed
            self.location.get_panel_meta_data()
            assert self.location.arming_state.is_disarmed()

    def tests_get_panel_meta_data_none(self):
        """Test get_panel_meta_data() with an empty PanelMetadataAndStatus response."""
        responses = [RESPONSE_DISARMED_NONE]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start disarmed
            assert self.location.arming_state.is_disarmed()

            # first response gives empty status...should remain the same
            with pytest.raises(PartialResponseError):
                self.location.get_panel_meta_data()

            assert self.location.arming_state.is_disarmed()

    def tests_get_panel_meta_data_failed(self):
        """Test get_panel_meta_data() with an empty PanelMetadataAndStatus response."""
        responses = [RESPONSE_FEATURE_NOT_SUPPORTED]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start disarmed
            assert self.location.arming_state.is_disarmed()

            with pytest.raises(FeatureNotSupportedError):
                self.location.get_panel_meta_data()



"""Test total_connect_client."""

from unittest.mock import patch

from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    RESPONSE_ARMED_AWAY,
    TCC_REQUEST_METHOD,
    # these are tested in test_client_arm_disarm for now
    # RESPONSE_ARMED_STAY,
    # RESPONSE_ARMED_STAY_NIGHT,
    # RESPONSE_DISARMED,
    # RESPONSE_FEATURE_NOT_SUPPORTED,
)

from total_connect_client.const import _ResultCode
from total_connect_client.client import ArmingHelper

# from total_connect_client.exceptions import AuthenticationError, BadResultCodeError


RESPONSE_ARM_SUCCESS = {
    "ResultCode": _ResultCode.ARM_SUCCESS.value,
    "ResultData": "testing arm success",
}

# returned when a zone is faulted
RESPONSE_ARM_FAILED = {
    "ResultCode": _ResultCode.COMMAND_FAILED.value,
    "ResultData": "testing arm failed",
}
RESPONSE_DISARM_FAILED = {
    "ResultCode": _ResultCode.COMMAND_FAILED.value,
    "ResultData": "testing disarm failed",
}

# appears to be for a bad/wrong code
RESPONSE_USER_CODE_INVALID = {
    "ResultCode": _ResultCode.USER_CODE_INVALID.value,
    "ResultData": "testing user code invalid",
}

# appears to be for a code entered for a wrong device/location
RESPONSE_USER_CODE_UNAVAILABLE = {
    "ResultCode": _ResultCode.USER_CODE_UNAVAILABLE.value,
    "ResultData": "testing user code unavailable",
}

RESPONSE_SUCCESS = {
    "ResultCode": _ResultCode.SUCCESS.value,
    "ResultData": "testing success",
}


def tests_arm_away():
    """Test arm away."""
    # first test with no issues
    client = create_client()
    location = client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]]

    responses = [RESPONSE_ARM_SUCCESS, RESPONSE_ARMED_AWAY]
    with patch(TCC_REQUEST_METHOD, side_effect=responses):
        ArmingHelper(location).arm_away()

        # confirm armed_away
        location.get_panel_meta_data()
        assert location.arming_state.is_armed_away()
        assert location.partitions["1"].arming_state.is_armed_away()



"""Experimental API calls to find new capability."""

import logging
import sys
import getpass
from pprint import pprint

from total_connect_client.client import TotalConnectClient

logging.basicConfig(filename="test.log", level=logging.DEBUG)

if len(sys.argv) < 2 or len(sys.argv) > 3:
    print("usage:  python3 -m total_connect_client username [password]\n")
    sys.exit(1)


USERNAME = sys.argv[1]
if len(sys.argv) == 2:
    PASSWORD = getpass.getpass()
else:
    PASSWORD = sys.argv[2]

TC = TotalConnectClient(USERNAME, PASSWORD)


def request_location(request_name, location_id):
    """Request with just location and token."""
    print(f"\n{request_name}: {location_id}")
    try:
        result = TC.request(request_name, (TC.token, location_id))
        pprint(result)
    except Exception as err:
        print(err)


def request_device(request_name, device_id):
    """Request with just device and token."""
    print(f"\n{request_name}: {device_id}")
    try:
        result = TC.request(request_name, (TC.token, device_id))
        pprint(result)
    except Exception as err:
        print(err)


print("GetDoorBellClientID")
try:
    result = TC.request("GetDoorBellClientID", (TC.token,))
    pprint(result)
except Exception as err:
    print(err)

for location_id, location in TC.locations.items():
    print(f"{location}")

    # camera
    location.get_cameras()
    location.get_video()
    request_location("GetAllRSIDeviceStatus", location_id)
    request_location("GetPartnerCameraStatus", location_id)

    # thermostat
    # request_location("GetWiFiThermostatLocations", location_id)

    # lock
    # request_location("GetWiFiLockLocations", location_id)

    # automation / scenes
    # request_location("GetSmartActionConfiguration", location_id)
    # request_location("GetSmartSceneConfiguration", location_id)
    # request_location("GetSmartSceneList", location_id)

    print(f"\n\nGetSmartActionList: {location_id} with no array")
    try:
        result = TC.request("GetSmartActionList", (TC.token, location_id, None, True))
        pprint(result)
    except Exception as err:
        print(err)
    print(f"\n\nGetSmartActionList: {location_id} with empty array")
    try:
        result = TC.request("GetSmartActionList", (TC.token, location_id, [], True))
        pprint(result)
    except Exception as err:
        print(err)

    for device_id, device in location.devices.items():
        print(f"{device}")

        # request_device("GetAutomationDeviceStatus", device_id)
        # request_device("GetAutomationDeviceStatusExV1", device_id)
        request_device("GetAllAutomationDeviceStatusExV1", device_id)
        # hide scene stuff for now
        # request_device("GetSceneList", device_id)

        if device.is_doorbell():
            print(
                f"\n\nGetWiFiDoorBellDeviceDetailsEx: location {location_id} device {device_id}"
            )
            try:
                result = TC.request(
                    "GetWiFiDoorBellDeviceDetailsEx", (TC.token, location_id, device_id)
                )
                pprint(result)
            except Exception as err:
                print(err)

            """
            print(
                f"\n\nGetWiFiDoorBellDeviceDiagnosticDetailsEx: location {location_id} device {device_id}"
            )
            try:
                result = TC.request(
                    "GetWiFiDoorBellDeviceDiagnosticDetailsEx",
                    (TC.token, location_id, device_id),
                )
                pprint(result)
            except Exception as err:
                print(err)

            print(
                f"\n\nGetWiFiDoorBellSettings: location {location_id} device {device._doorbell_info['PartnerDeviceID']}"
            )
            try:
                result = TC.request(
                    "GetWiFiDoorBellSettings",
                    (TC.token, location_id, device._doorbell_info["PartnerDeviceID"]),
                )
                pprint(result)
            except Exception as err:
                print(err)
            """



"""Test total_connect_client."""

import unittest
from unittest.mock import patch

import pytest
from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    RESPONSE_ARMED_AWAY,
    RESPONSE_ARMED_STAY,
    RESPONSE_ARMED_STAY_NIGHT,
    RESPONSE_DISARMED,
)

from total_connect_client import ArmingHelper
from total_connect_client.const import ArmingState, _ResultCode
from total_connect_client.exceptions import BadResultCodeError, UsercodeInvalid, UsercodeUnavailable

TCC_REQUEST_METHOD = "total_connect_client.client.TotalConnectClient.request"

RESPONSE_ARM_SUCCESS = {
    "ResultCode": _ResultCode.ARM_SUCCESS.value,
    "ResultData": "testing arm success",
}
RESPONSE_DISARM_SUCCESS = {
    "ResultCode": _ResultCode.DISARM_SUCCESS.value,
    "ResultData": "testing disarm success",
}

# returned when a zone is faulted
RESPONSE_ARM_FAILED = {
    "ResultCode": _ResultCode.COMMAND_FAILED.value,
    "ResultData": "testing arm failed",
}
RESPONSE_DISARM_FAILED = {
    "ResultCode": _ResultCode.COMMAND_FAILED.value,
    "ResultData": "testing disarm failed",
}

# appears to be for a bad/wrong code
RESPONSE_USER_CODE_INVALID = {
    "ResultCode": _ResultCode.USER_CODE_INVALID.value,
    "ResultData": "testing user code invalid",
}

# appears to be for a code entered for a wrong device/location
RESPONSE_USER_CODE_UNAVAILABLE = {
    "ResultCode": _ResultCode.USER_CODE_UNAVAILABLE.value,
    "ResultData": "testing user code unavailable",
}

RESPONSE_SUCCESS = {
    "ResultCode": _ResultCode.SUCCESS.value,
    "ResultData": "testing success",
}


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]

    def tests_arm(self):
        def run(error, responses, armit, armingstate):
            client = create_client()
            location = client.locations[self.location_id]
            with patch(TCC_REQUEST_METHOD, side_effect=responses):
                if error:
                    with pytest.raises(error):
                        armit(location)
                else:
                    armit(location)
                location.get_panel_meta_data()
                assert location.arming_state == armingstate

        def runall(success_response, armit, success_armingstate):
            # first test is all good
            run(
                None,
                [RESPONSE_ARM_SUCCESS, success_response],
                armit,
                success_armingstate,
            )
            for failresponse in (
                RESPONSE_ARM_FAILED,  # "zone faulted"
            ):
                run(
                    BadResultCodeError,
                    [failresponse, RESPONSE_DISARMED],
                    armit,
                    ArmingState.DISARMED,
                )
            # last test has 'unavailable' usercode
            run(
                UsercodeUnavailable,
                [RESPONSE_USER_CODE_UNAVAILABLE, RESPONSE_DISARMED],
                armit,
                ArmingState.DISARMED,
            )

        runall(
            RESPONSE_ARMED_AWAY,
            lambda loc: ArmingHelper(loc).arm_away(),
            ArmingState.ARMED_AWAY,
        )
        # this test claims to be testing arm away instant but uses the
        # arm away response
        runall(
            RESPONSE_ARMED_AWAY,
            lambda loc: ArmingHelper(loc).arm_away_instant(),
            ArmingState.ARMED_AWAY,
        )
        runall(
            RESPONSE_ARMED_STAY,
            lambda loc: ArmingHelper(loc).arm_stay(),
            ArmingState.ARMED_STAY,
        )
        # this test claims to be testing arm stay instant but uses the
        # arm stay response
        runall(
            RESPONSE_ARMED_STAY,
            lambda loc: ArmingHelper(loc).arm_stay_instant(),
            ArmingState.ARMED_STAY,
        )
        runall(
            RESPONSE_ARMED_STAY_NIGHT,
            lambda loc: ArmingHelper(loc).arm_stay_night(),
            ArmingState.ARMED_STAY_NIGHT,
        )

    def tests_disarm(self):
        """Test disarm."""
        # first test with no issues
        client = create_client()
        location = client.locations[self.location_id]
        responses = [
            RESPONSE_ARM_SUCCESS,
            RESPONSE_ARMED_AWAY,
            RESPONSE_DISARM_SUCCESS,
            RESPONSE_DISARMED,
        ]
        with patch(TCC_REQUEST_METHOD, side_effect=responses):
            # arm the system and confirm armed_away
            ArmingHelper(location).arm_away()
            location.get_panel_meta_data()
            assert location.arming_state.is_armed_away()

            # now disarm
            ArmingHelper(location).disarm()
            location.get_panel_meta_data()
            assert location.arming_state.is_disarmed()

    def tests_disarm_command_failed(self):
        """Test disarm with command failed."""
        # first test with no issues
        client = create_client()
        location = client.locations[self.location_id]
        responses = [
            RESPONSE_ARM_SUCCESS,
            RESPONSE_ARMED_AWAY,
            RESPONSE_DISARM_FAILED,
            RESPONSE_ARMED_AWAY,
        ]
        with patch(TCC_REQUEST_METHOD, side_effect=responses):
            # arm the system and confirm armed_away
            ArmingHelper(location).arm_away()
            location.get_panel_meta_data()
            assert location.arming_state.is_armed_away()

            # now disarm, should fail
            with pytest.raises(BadResultCodeError):
                ArmingHelper(location).disarm()

            # should still be armed_away
            location.get_panel_meta_data()
            assert location.arming_state.is_armed_away()

    def tests_disarm_user_code_invalid(self):
        """Test disarm with invalid user code."""
        client = create_client()
        location = client.locations[self.location_id]
        responses = [
            RESPONSE_ARM_SUCCESS,
            RESPONSE_ARMED_AWAY,
            RESPONSE_USER_CODE_INVALID,
            RESPONSE_ARMED_AWAY,
        ]
        with patch(TCC_REQUEST_METHOD, side_effect=responses):
            # arm the system and confirm armed_away
            ArmingHelper(location).arm_away()
            location.get_panel_meta_data()
            assert location.arming_state.is_armed_away()

            with pytest.raises(UsercodeInvalid):
                ArmingHelper(location).disarm()

            # should still be armed_away when disarming fails
            location.get_panel_meta_data()
            assert location.arming_state.is_armed_away()

    def tests_disarm_disarmed(self):
        """Test attempt to disarm an already disarmed system."""
        # Did this once on my Lynx 7000 and it gave result code SUCCESS

        client = create_client()
        location = client.locations[self.location_id]
        responses = [RESPONSE_DISARMED, RESPONSE_SUCCESS, RESPONSE_DISARMED]
        with patch(TCC_REQUEST_METHOD, side_effect=responses):
            location.get_panel_meta_data()
            assert location.arming_state.is_disarmed()

            # now disarm
            ArmingHelper(location).disarm()
            location.get_panel_meta_data()
            assert location.arming_state.is_disarmed()



"""Test total_connect_client request()."""

import unittest
from unittest.mock import patch
from zeep.exceptions import Fault as ZeepFault

import pytest
import requests_mock
from const import (
    MAX_RETRY_ATTEMPTS,
    LOCATION_INFO_BASIC_NORMAL,
    RESPONSE_ARMED_AWAY,
    RESPONSE_SESSION_DETAILS,
    RESPONSE_CONNECTION_ERROR,
    RESPONSE_DISARMED,
    RESPONSE_FAILED_TO_CONNECT,
    RESPONSE_GET_ZONE_DETAILS_SUCCESS,
    RESPONSE_INVALID_SESSION,
    RESPONSE_PARTITION_DETAILS,
    RESPONSE_UNKNOWN,
    HTTP_RESPONSE_TOKEN,
    HTTP_RESPONSE_TOKEN_2,
    HTTP_RESPONSE_BAD_USER_OR_PASSWORD,
    HTTP_RESPONSE_REFRESH_TOKEN_FAILED,
    SESSION_ID,
    SESSION_ID_2,
    TOKEN_EXPIRATION_TIME
)

from total_connect_client.const import ArmType
from total_connect_client.client import TotalConnectClient
from total_connect_client.exceptions import AuthenticationError, BadResultCodeError, ServiceUnavailable

PATCH_EVAL = "total_connect_client.client.TotalConnectClient._send_one_request"


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient request()."""

    def setUp(self):
        """Test setup."""
        self.client = None
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_request_init(self):
        """Test normal init sequence with no problems."""
        serialize_responses = [
            RESPONSE_SESSION_DETAILS,
            RESPONSE_PARTITION_DETAILS,
            RESPONSE_GET_ZONE_DETAILS_SUCCESS,
            RESPONSE_DISARMED,
        ]

        with patch("zeep.Client"), patch(
            PATCH_EVAL, side_effect=serialize_responses
        ) as mock_request:
            client = TotalConnectClient("username", "password", usercodes=None)
            assert mock_request.call_count == 1
            if client.locations:  # force client to fetch them
                pass
            assert mock_request.call_count == 4
            assert client.is_logged_in() is True

    def tests_request_init_bad_user_or_password(self):
        """Test init sequence with no a bad password."""
        with requests_mock.Mocker(real_http=True) as rm:
            rm.post(
                TotalConnectClient.TOKEN_ENDPOINT,
                json=HTTP_RESPONSE_BAD_USER_OR_PASSWORD,
                status_code=403
            )
            with pytest.raises(AuthenticationError):
                TotalConnectClient("username", "password", usercodes=None)

    def tests_request_init_failed_to_connect(self):
        """Test init sequence when it fails to connect."""
        serialize_responses = [
            RESPONSE_FAILED_TO_CONNECT for x in range(MAX_RETRY_ATTEMPTS)
        ]

        with patch("zeep.Client"), patch("time.sleep", autospec=True), patch(
            PATCH_EVAL, side_effect=serialize_responses
        ) as mock_request, pytest.raises(Exception) as exc:
            client = TotalConnectClient(
                "username", "password", usercodes=None, retry_delay=0
            )
            assert mock_request.call_count == MAX_RETRY_ATTEMPTS
            assert client.is_logged_in() is False
            expected = "total-connect-client could not execute request.  Maximum attempts tried."
            assert str(exc.value) == expected

    def tests_request_connection_error(self):
        """Test a connection error."""
        serialize_responses = [
            RESPONSE_CONNECTION_ERROR for x in range(MAX_RETRY_ATTEMPTS)
        ]

        with patch("zeep.Client"), patch("time.sleep", autospec=True), patch(
            "zeep.helpers.serialize_object", side_effect=serialize_responses
        ) as mock_request, pytest.raises(Exception) as exc:
            client = TotalConnectClient(
                "username", "password", usercodes=None, retry_delay=0
            )
            assert mock_request.call_count == MAX_RETRY_ATTEMPTS
            assert client.is_logged_in() is False
            expected = "total-connect-client could not execute request.  Maximum attempts tried."
            assert str(exc.value) == expected

    def tests_request_zeep_error(self):
        """Test a Zeep error."""

        serialize_responses = [ZeepFault("test") for x in range(MAX_RETRY_ATTEMPTS)]
        with patch("zeep.Client"), patch("time.sleep", autospec=True), patch(
            "zeep.helpers.serialize_object", side_effect=serialize_responses
        ) as mock_request, pytest.raises(ServiceUnavailable):
            client = TotalConnectClient(
                "username", "password", usercodes=None, retry_delay=0
            )
            assert mock_request.call_count == MAX_RETRY_ATTEMPTS
            assert client.is_logged_in() is False

    def tests_request_invalid_session(self):
        """Test an invalid session, which is when the session times out."""
        # First four responses set up 'normal' session
        # Call to client.arm_away() will first get an invalid session,
        # which will trigger client.authenticate() before completing the arm_away()
        serialize_responses = [
            RESPONSE_SESSION_DETAILS,
            RESPONSE_PARTITION_DETAILS,
            RESPONSE_GET_ZONE_DETAILS_SUCCESS,
            RESPONSE_DISARMED,
            RESPONSE_INVALID_SESSION,
            RESPONSE_ARMED_AWAY,
        ]

        token_responses = [
            {"json": HTTP_RESPONSE_TOKEN, "status_code": 200},
            {"json": HTTP_RESPONSE_TOKEN_2, "status_code": 200},
        ]

        with patch("zeep.Client"), patch(
            PATCH_EVAL, side_effect=serialize_responses
        ) as mock_request, requests_mock.Mocker(real_http=True) as rm:
            rm.post(TotalConnectClient.TOKEN_ENDPOINT, token_responses)
            client = TotalConnectClient("username", "password", usercodes=None)
            assert mock_request.call_count == 1
            if client.locations:  # force client to fetch them
                pass
            assert mock_request.call_count == 4
            assert client.is_logged_in() is True

            assert client.token == SESSION_ID
            # now try to arm away
            # the invalid session will trigger a new authenticate()...
            # which should result in a new token
            client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]].arm(ArmType.AWAY)
            assert mock_request.call_count == 6
            assert client.token == SESSION_ID_2

    def tests_request_unknown_result_code(self):
        """Test an unknown result code."""
        serialize_responses = [
            RESPONSE_UNKNOWN,
        ]

        with patch("zeep.Client"), patch(
            PATCH_EVAL, side_effect=serialize_responses
        ) as mock_request:
            with pytest.raises(BadResultCodeError):
                TotalConnectClient("username", "password", usercodes=None)
            assert mock_request.call_count == 1

    def tests_empty_response_code(self):
        """Test an empty response code."""
        # see issue #228
        from total_connect_client.const import _ResultCode
        with pytest.raises(ServiceUnavailable):
            _ResultCode.from_response(None)

    def tests_refresh_session(self):
        """Test various scenarious around session refreshing. First the session expires and is
        successfully refreshed, then the session expires and the refresh fails, forcing a full
        reauthentication"""
        # First four responses set up 'normal' session
        # Call to client.arm_away() will first require a session refresh
        # Call to client.disarm() will first require a session refresh
        serialize_responses = [
            RESPONSE_SESSION_DETAILS,
            RESPONSE_PARTITION_DETAILS,
            RESPONSE_GET_ZONE_DETAILS_SUCCESS,
            RESPONSE_DISARMED,
            RESPONSE_ARMED_AWAY,
            RESPONSE_DISARMED
        ]

        # Mocked values for 'time.monotonic'
        mock_times = [
            # Time for initial authentication
            0,

            # Session still valid before RESPONSE_SESSION_DETAILS
            TOKEN_EXPIRATION_TIME - 5,

            # Session still valid before RESPONSE_PARTITION_DETAILS
            TOKEN_EXPIRATION_TIME - 4,

            # Session still valid before RESPONSE_GET_ZONE_DETAILS_SUCCESS
            TOKEN_EXPIRATION_TIME - 3,

            # Session still valid before RESPONSE_DISARMED
            TOKEN_EXPIRATION_TIME - 2,

            # Session expires before RESPONSE_ARMED_AWAY
            TOKEN_EXPIRATION_TIME,

            # Time for session refresh
            TOKEN_EXPIRATION_TIME,

            # Session expires before RESPONSE_DISARMED
            TOKEN_EXPIRATION_TIME * 2,

            # Time for reauthentication
            TOKEN_EXPIRATION_TIME * 2,

            # Session now valid for RESPONSE_DISARMED
            TOKEN_EXPIRATION_TIME * 2 + 1,
        ]

        token_responses = [
            # Successful response for initial authentication
            {"json": HTTP_RESPONSE_TOKEN, "status_code": 200},

            # Successful response for session refresh
            {"json": HTTP_RESPONSE_TOKEN, "status_code": 200},

            # Failed response for refresh
            {"json": HTTP_RESPONSE_REFRESH_TOKEN_FAILED, "status_code": 400},

            # Successful response for reauthentication
            {"json": HTTP_RESPONSE_TOKEN_2, "status_code": 200},
        ]

        with patch("zeep.Client"), patch(
            PATCH_EVAL, side_effect=serialize_responses
        ) as mock_request, patch(
            "time.monotonic", side_effect=mock_times
        ), requests_mock.Mocker(real_http=True) as rm:
            rm.post(TotalConnectClient.TOKEN_ENDPOINT, token_responses)

            client = TotalConnectClient("username", "password", usercodes=None)
            assert mock_request.call_count == 1
            if client.locations:  # force client to fetch them
                pass
            assert mock_request.call_count == 4
            assert client.is_logged_in() is True
            assert client.token == SESSION_ID

            # The session should come up as expired due to the current time.
            # The client refreshes the session and should get the same session id.
            client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]].arm(ArmType.AWAY)
            assert mock_request.call_count == 5
            assert client.token == SESSION_ID

            # Check that we made a refresh session request
            last_http_request = rm.request_history[-1]
            assert last_http_request.url == TotalConnectClient.TOKEN_ENDPOINT
            assert "grant_type=refresh_token" in last_http_request.text

            # The session should come up as expired again due to the current time.
            # The session refresh fails so we reauthenticate and get a new session.
            client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]].disarm()
            assert mock_request.call_count == 6
            assert client.token == SESSION_ID_2

            # Check that we made an authentication request
            last_http_request = rm.request_history[-1]
            assert last_http_request.url == TotalConnectClient.TOKEN_ENDPOINT
            assert "grant_type=password" in last_http_request.text



"""Test client authentication."""

import unittest
from unittest.mock import patch

import pytest
import requests_mock
from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    HTTP_RESPONSE_BAD_USER_OR_PASSWORD,
)

from total_connect_client.const import _ResultCode
from total_connect_client.client import TotalConnectClient
from total_connect_client.exceptions import AuthenticationError

RESPONSE_SUCCESS = {
    "ResultCode": _ResultCode.SUCCESS.value,
    "ResultData": "None",
    "SessionID": "123456890",
}


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.client = create_client()
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_logout(self):
        """Test logout."""
        responses = [
            RESPONSE_SUCCESS,
        ]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            assert self.client.is_logged_in() is True
            self.client.log_out()
            assert self.client.is_logged_in() is False

            # succeeds because we are logged out
            self.client.log_out()

    def tests_authenticate(self):
        """Test authenticate()."""
        responses = [
            RESPONSE_SUCCESS,
            RESPONSE_SUCCESS,
        ]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ), patch(
            "total_connect_client.client.TotalConnectClient._make_locations",
            return_value=["fakelocations"],
        ):
            # ensure we start logged out (first SUCCESS)
            self.client.log_out()
            assert self.client.is_logged_in() is False

            # success (second SUCCESS)
            self.client.authenticate()
            assert self.client.is_logged_in() is True

            # bad user or pass
            with requests_mock.Mocker(real_http=True) as rm, pytest.raises(AuthenticationError):
                rm.post(
                    TotalConnectClient.TOKEN_ENDPOINT,
                    json=HTTP_RESPONSE_BAD_USER_OR_PASSWORD,
                    status_code=403
                )
                self.client.authenticate()
            assert self.client.is_logged_in() is False

            # authentication failed
            with pytest.raises(AuthenticationError):
                self.client.authenticate()
            assert self.client.is_logged_in() is False



"""Test client.validate_usercode()."""

import unittest
from unittest.mock import patch

import pytest
from common import create_client
from const import LOCATION_INFO_BASIC_NORMAL

from total_connect_client.const import _ResultCode
from total_connect_client.exceptions import BadResultCodeError

RESPONSE_VALID = {
    "ResultCode": _ResultCode.SUCCESS.value,
    "ResultData": "None",
}

RESPONSE_INVALID = {
    "ResultCode": _ResultCode.USER_CODE_INVALID.value,
    "ResultData": "None",
}

RESPONSE_UNAVAILABLE = {
    "ResultCode": _ResultCode.USER_CODE_UNAVAILABLE.value,
    "ResultData": "None",
}

# random unknown response
RESPONSE_FAILED = {
    "ResultCode": _ResultCode.COMMAND_FAILED.value,
    "ResultData": "None",
}


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.client = create_client()
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_validate_usercode(self):
        """Test validate_usercode() with a valid code."""
        responses = [
            RESPONSE_VALID,
            RESPONSE_INVALID,
            RESPONSE_UNAVAILABLE,
            RESPONSE_FAILED,
        ]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # valid
            assert self.client.validate_usercode("1", "1234") is True
            # invalid
            assert self.client.validate_usercode("1", "1234") is False
            # unavailable
            assert self.client.validate_usercode("1", "1234") is False
            # other random failure
            with pytest.raises(BadResultCodeError):
                self.client.validate_usercode("1", "1234")



"""Total Connect Location."""

import logging
from typing import Dict, Any

from .const import PROJECT_URL, ArmingState, ArmType, _ResultCode
from .device import TotalConnectDevice
from .exceptions import (
    FeatureNotSupportedError,
    PartialResponseError,
    TotalConnectError,
)
from .partition import TotalConnectPartition
from .zone import TotalConnectZone, ZoneStatus

DEFAULT_USERCODE = "-1"

LOGGER = logging.getLogger(__name__)


class TotalConnectLocation:
    """TotalConnectLocation class."""

    def __init__(self, location_info_basic: Dict[str, Any], parent) -> None:
        """Initialize based on a 'LocationInfoBasic'."""
        self.location_id = location_info_basic["LocationID"]
        self.location_name: str = location_info_basic["LocationName"]
        self._photo_url: str = location_info_basic["PhotoURL"]
        self._module_flags = dict(
            x.split("=") for x in location_info_basic["LocationModuleFlags"].split(",")
        )
        self.security_device_id: str = location_info_basic["SecurityDeviceID"]
        self.parent = parent
        self.ac_loss = None
        self.low_battery = None
        self.cover_tampered = None
        self.last_updated_timestamp_ticks = None
        self.configuration_sequence_number = None
        self.arming_state: ArmingState = ArmingState.UNKNOWN
        self.partitions: Dict[Any, TotalConnectPartition] = {}
        self._partition_list: Dict[str, list[Any]] = {"int": []}
        self.zones: Dict[Any, TotalConnectZone] = {}
        self.usercode: str = DEFAULT_USERCODE
        self.auto_bypass_low_battery: bool = False
        self._sync_job_id = None
        self._sync_job_state: int = 0

        dib = (location_info_basic.get("DeviceList") or {}).get("DeviceInfoBasic")
        tcdevs = [TotalConnectDevice(d) for d in (dib or {})]
        self.devices = {tcdev.deviceid: tcdev for tcdev in tcdevs}

    def __str__(self) -> str:
        """Return a text string that is printable."""
        data = (
            f"LOCATION {self.location_id} - {self.location_name}\n\n"
            f"PhotoURL: {self._photo_url}\n"
            f"SecurityDeviceID: {self.security_device_id}\n"
            f"AcLoss: {self.ac_loss}\n"
            f"LowBattery: {self.low_battery}\n"
            f"IsCoverTampered: {self.cover_tampered}\n"
            f"{self.arming_state}\n"
            f"LocationModuleFlags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + "\n"

        devices = f"DEVICES: {len(self.devices)}\n\n"
        for device in self.devices:
            devices = devices + str(self.devices[device]) + "\n"

        partitions = f"PARTITIONS: {len(self.partitions)}\n\n"
        for status in self.partitions.values():
            partitions += str(status) + "\n"

        zones = f"ZONES: {len(self.zones)}\n\n"
        for status in self.zones.values():
            zones += str(status)

        return data + devices + partitions + zones

    def get_panel_meta_data(self) -> None:
        """Get all meta data about the alarm panel."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=GetPanelMetaDataAndFullStatus
        result = self.parent.request(
            "GetPanelMetaDataAndFullStatusEx_V2",
            (
                # to speed this up we could replace the first zero with
                # the most recent ConfigurationSequenceNumber and the
                # second with LastUpdatedTimestampTicks
                self.parent.token,
                self.location_id,
                0,
                0,
                self._partition_list,
            ),
        )
        self.parent.raise_for_resultcode(result)

        self._update_status(result)
        self._update_partitions(result)
        self._update_zones(result)

    def get_zone_details(self) -> None:
        """Get Zone details."""
        result = self.parent.request(
            "GetZonesListInStateEx_V1",
            (
                # 0 is the ListIdentifierID, whatever that might be
                self.parent.token,
                self.location_id,
                self._partition_list,
                0,
            ),
        )

        try:
            self.parent.raise_for_resultcode(result)
            self._update_zone_details(result)
        except FeatureNotSupportedError:
            LOGGER.warning(
                "getting Zone Details is a feature not supported by "
                "your Total Connect account or hardware"
            )

    def get_partition_details(self) -> None:
        """Get partition details for this location."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=GetPartitionsDetails

        result = self.parent.request(
            "GetPartitionsDetails",
            (self.parent.token, self.location_id, self.security_device_id),
        )
        try:
            self.parent.raise_for_resultcode(result)
        except TotalConnectError:
            LOGGER.error(
                f"Could not get partition details for "
                f"device {self.security_device_id} at "
                f"location {self.location_id}."
                f"ResultCode: {result['ResultCode']}. "
                f"ResultData: {result['ResultData']}"
            )
            raise

        partition_details = ((result or {}).get("PartitionsInfoList") or {}).get(
            "PartitionDetails"
        )
        if not partition_details:
            raise PartialResponseError("no PartitionDetails", result)

        new_partition_list = []
        for partition in partition_details:
            new_partition = TotalConnectPartition(partition, self)
            self.partitions[new_partition.partitionid] = new_partition
            new_partition_list.append(new_partition.partitionid)

        self._partition_list = {"int": new_partition_list}

    def is_low_battery(self) -> bool:
        """Return true if low battery."""
        return self.low_battery is True

    def is_ac_loss(self) -> bool:
        """Return true if AC loss."""
        return self.ac_loss is True

    def is_cover_tampered(self) -> bool:
        """Return true if cover is tampered."""
        return self.cover_tampered is True

    def set_usercode(self, usercode: str) -> bool:
        """Set the usercode. Return true if successful."""
        if self.parent.validate_usercode(self.security_device_id, usercode):
            self.usercode = usercode
            return True
        return False

    def _build_partition_list(self, partition_id:str="") -> Dict[str, list[Any]]:
        """Build a list of partitions to use for arming/disarming."""
        if not partition_id:
            return self._partition_list

        if partition_id not in self.partitions:
            raise TotalConnectError(
                f"Partition {partition_id} does not exist "
                f"at location {self.location_id}"
            )
        return {"int": [partition_id]}

    def arm(self, arm_type: ArmType, partition_id:str="", usercode: str = "") -> None:
        """Arm the given partition.

        If no partition is given, arm all partitions.
        If no usercode given, use stored value."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=ArmSecuritySystemPartitionsV1
        assert isinstance(arm_type, ArmType)
        partition_list = self._build_partition_list(partition_id)
        usercode = usercode or self.usercode

        result = self.parent.request(
            "ArmSecuritySystemPartitionsV1",
            (
                self.parent.token,
                self.location_id,
                self.security_device_id,
                arm_type.value,
                usercode,
                partition_list,
            ),
        )
        if _ResultCode.from_response(result) == _ResultCode.COMMAND_FAILED:
            LOGGER.warning("could not arm system; is a zone faulted?")
        self.parent.raise_for_resultcode(result)
        LOGGER.info(
            f"ARMED({arm_type}) partitions {partition_list} at {self.location_id}"
        )

    def disarm(self, partition_id:str="", usercode: str = "") -> None:
        """Disarm the system."""
        # if no partition is given, disarm all partitions
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=ArmSecuritySystemPartitionsV1
        partition_list = self._build_partition_list(partition_id)
        usercode = usercode or self.usercode

        result = self.parent.request(
            "DisarmSecuritySystemPartitionsV1",
            (
                self.parent.token,
                self.location_id,
                self.security_device_id,
                usercode,
                partition_list,
            ),
        )
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"DISARMED partitions {partition_list} at {self.location_id}")

    def zone_bypass(self, zone_id: int) -> None:
        """Bypass a zone."""
        result = self.parent.request(
            "Bypass",
            (
                self.parent.token,
                self.location_id,
                self.security_device_id,
                zone_id,
                self.usercode,
            ),
        )
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"BYPASSED {zone_id} at {self.location_id}")
        self.zones[zone_id]._mark_as_bypassed()

    def zone_bypass_all(self) -> None:
        """Bypass all faulted zones."""
        faulted_zones = []
        for zone_id, zone in self.zones.items():
            if zone.is_faulted():
                faulted_zones.append(zone_id)

        if faulted_zones:
            zone_list = {"int": faulted_zones}

            result = self.parent.request(
                "BypassAll",
                (
                    self.parent.token,
                    self.location_id,
                    self.security_device_id,
                    zone_list,
                    self.usercode,
                ),
            )
            self.parent.raise_for_resultcode(result)
            LOGGER.info(f"BYPASSED all zones at location {self.location_id}")

    def clear_bypass(self) -> None:
        """Clear all bypassed zones."""
        self.disarm()

    def zone_status(self, zone_id: int) -> ZoneStatus:
        """Get status of a zone."""
        zone = self.zones.get(zone_id)
        if not zone:
            raise TotalConnectError(f"zone {zone_id} does not exist")
        return zone.status

    def arm_custom(self, arm_type: ArmType) -> Dict[str, Any]:
        """NOT OPERATIONAL YET.
        Arm custom the system.  Return true if successful.
        """
        zones_list = {}
        zones_list[0] = {"ZoneID": "12", "ByPass": False, "ZoneStatus": 0}
        settings = {"ArmMode": "1", "ArmDelay": "5", "ZonesList": zones_list}

        result = self.parent.request(
            "CustomArmSecuritySystem",
            (
                self.parent.token,
                self.location_id,
                self.security_device_id,
                arm_type.value,
                self.usercode,
                settings,
            ),
        )
        self.parent.raise_for_resultcode(result)
        # TODO: returning the raw result is not right
        return result

    def get_custom_arm_settings(self) -> Dict[str, Any]:
        """NOT OPERATIONAL YET.
        Get custom arm settings.
        """
        result = self.parent.request(
            "GetCustomArmSettings",
            (self.parent.token, self.location_id, self.security_device_id),
        )
        self.parent.raise_for_resultcode(result)
        # TODO: returning the raw result is not right
        return result

    def _update_zone_details(self, result: Dict[str, Any]) -> None:
        """
        Update from GetZonesListInStateEx_V1.

        ZoneStatusInfoWithPartitionId provides additional info for setting up zones.
        If we used TotalConnectZone._update() it would overwrite missing data with None.
        """
        zone_info = ((result.get("ZoneStatus") or {}).get("Zones") or {}).get(
            "ZoneStatusInfoWithPartitionId"
        )
        if not zone_info:
            LOGGER.warning(
                "No zones found when starting TotalConnect. Try to sync your panel using the TotalConnect app or website."
            )
            LOGGER.debug(f"_update_zone_details result: {result}")
        else:
            for zonedata in zone_info:
                self.zones[zonedata["ZoneID"]] = TotalConnectZone(zonedata, self)

    def _update_status(self, result: Dict[str, Any]) -> None:
        """Update from result."""
        data = (result or {}).get("PanelMetadataAndStatus")
        if not data:
            raise PartialResponseError("no PanelMetadataAndStatus", result)

        self.ac_loss = data.get("IsInACLoss")
        self.low_battery = data.get("IsInLowBattery")
        self.cover_tampered = data.get("IsCoverTampered")
        self.last_updated_timestamp_ticks = data.get("LastUpdatedTimestampTicks")
        self.configuration_sequence_number = data.get("ConfigurationSequenceNumber")

        astate = result.get("ArmingState")
        if not astate:
            raise PartialResponseError("no ArmingState", result)
        try:
            self.arming_state = ArmingState(astate)
        except ValueError:
            LOGGER.error(
                f"unknown location ArmingState {astate} in {result}: please report at {PROJECT_URL}/issues"
            )
            raise TotalConnectError(
                f"unknown location ArmingState {astate} in {result}"
            ) from None

    def _update_partitions(self, result: Dict[str, Any]) -> None:
        """Update partition info from Partitions."""
        pinfo = (
            (result.get("PanelMetadataAndStatus") or {}).get("Partitions") or {}
        ).get("PartitionInfo")
        if not pinfo:
            raise PartialResponseError("no PartitionInfo", result)

        # loop through partitions and update
        # NOTE: do not use keys because they don't line up with PartitionID
        for partition in pinfo:
            if "PartitionID" not in partition:
                raise PartialResponseError("no PartitionID", result)
            partition_id = str(partition["PartitionID"])
            if partition_id in self.partitions:
                self.partitions[partition_id]._update(partition)
            else:
                LOGGER.warning(f"Update provided for unknown partion {partition_id}")

    def _update_zones(self, result: Dict[str, Any]) -> None:
        """Update zone info from ZoneInfo or ZoneInfoEx."""

        data = (result.get("PanelMetadataAndStatus") or {}).get("Zones")
        if not data:
            LOGGER.error(
                "no zones found: sync your panel using TotalConnect app or website"
            )
            # PartialResponseError would mean this is retryable without fixing
            # anything, and this needs fixing
            raise TotalConnectError("no zones found: panel sync required")

        zone_info = data.get("ZoneInfoEx") or data.get("ZoneInfo")
        if not zone_info:
            raise PartialResponseError("no ZoneInfoEx or ZoneInfo", result)
        for zonedata in zone_info:
            zid = (zonedata or {}).get("ZoneID")
            if not zid:
                raise PartialResponseError("no ZoneID", result)
            zone = self.zones.get(zid)
            if zone:
                zone._update(zonedata)
            else:
                zone = TotalConnectZone(zonedata, self)
                self.zones[zid] = zone

            if (
                zone.is_low_battery()
                and zone.can_be_bypassed
                and self.auto_bypass_low_battery
            ):
                self.zone_bypass(zid)

    def sync_panel(self) -> None:
        """Syncronize the panel with the TotalConnect server."""
        result = self.parent.request(
            "SynchronizeSecurityPanel",
            (self.parent.token, None, self.usercode, self.location_id, False),
        )
        self.parent.raise_for_resultcode(result)
        self._sync_job_id = result.get("JobID")
        # Successful request so assume state is in progress
        self._sync_job_state = 1
        LOGGER.info(f"Started sync of panel for location {self.location_id}")

    def get_sync_status(self) -> None:
        """Get panel sync status from the TotalConnect server."""
        result = self.parent.request(
            "GetSyncJobStatus", (self.parent.token, self._sync_job_id, self.location_id)
        )

        try:
            self.parent.raise_for_resultcode(result)
            job_state = result.get("JobState")
            if job_state == 1:
                LOGGER.info(f"Panel sync for location {self.location_id} in progress")
            elif job_state == 2:
                LOGGER.info(f"Panel sync for location {self.location_id} complete")
            else:
                LOGGER.warning(
                    f"Unknown panel sync status for location {self.location_id}"
                )
        except TotalConnectError:
            LOGGER.error(
                f"Could not get status of Sync Job with ID {self._sync_job_id}"
            )

    def get_cameras(self) -> None:
        """Get cameras for the location."""
        result = self.parent.request(
            "GetLocationAllCameraListEx", (self.parent.token, self.location_id)
        )
        self.parent.raise_for_resultcode(result)

        if "AccountAllCameraList" not in result:
            LOGGER.info(f"No cameras found for location {self.location_id}")
            return

        camera_list = result["AccountAllCameraList"]

        if "WiFiDoorbellList" in camera_list:
            self._get_doorbell(camera_list["WiFiDoorbellList"])

        if "UnicornList" in camera_list:
            self._get_unicorn(camera_list["UnicornList"])

    def _get_doorbell(self, data: Dict[str, Any]) -> None:
        """Find doorbell info."""
        if not data or "WiFiDoorbellsList" not in data:
            return

        doorbells = data["WiFiDoorbellsList"]
        if "WiFiDoorBellInfo" not in doorbells:
            return

        doorbell_list = doorbells["WiFiDoorBellInfo"]
        for doorbell in doorbell_list:
            id = doorbell["DeviceID"]
            if id in self.devices:
                self.devices[id].doorbell_info = doorbell

    def _get_unicorn(self, data: Dict[str, Any]) -> None:
        """Find uniforn info."""
        if not data or "UnicornList" not in data:
            return

        unicorns = data["UnicornsList"]
        if "UnicornInfo" not in unicorns:
            return

        unicorn_list = unicorns["UnicornInfo"]
        for unicorn in unicorn_list:
            id = unicorn["DeviceID"]
            if id in self.devices:
                self.devices[id].unicorn_info = unicorn

    def get_video(self) -> None:
        """Get video for the location."""
        result = self.parent.request(
            "GetVideoPIRLocationDeviceList", (self.parent.token, self.location_id)
        )
        self.parent.raise_for_resultcode(result)

        if "VideoPIRList" not in result:
            return

        if "VideoPIRInfo" not in result["VideoPIRList"]:
            return

        video_list = result["VideoPIRList"]["VideoPIRInfo"]
        for video in video_list:
            id = video["DeviceID"]
            if id in self.devices:
                self.devices[id].video_info = video



"""Test total_connect_client location."""

import unittest
from copy import deepcopy
from unittest.mock import Mock, patch

import pytest
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    METADATA_DISARMED,
    METADATA_DISARMED_LOW_BATTERY,
    RESPONSE_DISARMED,
    RESPONSE_GET_ZONE_DETAILS_SUCCESS,
)

from total_connect_client.const import ArmingState
from total_connect_client.exceptions import PartialResponseError, TotalConnectError
from total_connect_client.location import DEFAULT_USERCODE, TotalConnectLocation


class TestTotalConnectLocation(unittest.TestCase):
    """Test TotalConnectLocation."""

    def setUp(self):
        """Set up for location testing."""
        self.auto_bypass_low_battery = False
        self.location_normal = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, self)
        self.location_normal._update_status(deepcopy(RESPONSE_DISARMED))
        self.location_normal._update_partitions(deepcopy(RESPONSE_DISARMED))
        self.location_normal._update_zones(deepcopy(RESPONSE_DISARMED))

    def tearDown(self):
        """Tear down."""
        self.location_normal = None

    def tests_basic(self):
        """Test basic attributes were set properly."""
        self.assertTrue(
            self.location_normal.location_id == LOCATION_INFO_BASIC_NORMAL["LocationID"]
        )
        lname = LOCATION_INFO_BASIC_NORMAL["LocationName"]
        self.assertTrue(self.location_normal.location_name == lname)
        lsdid = LOCATION_INFO_BASIC_NORMAL["SecurityDeviceID"]
        self.assertTrue(self.location_normal.security_device_id == lsdid)

    def tests_panel(self):
        """Test panel attributes."""
        self.assertFalse(self.location_normal.is_low_battery())
        self.assertFalse(self.location_normal.is_ac_loss())
        self.assertFalse(self.location_normal.is_cover_tampered())

    def tests_status(self):
        """Normal zone."""
        self.assertFalse(self.location_normal.arming_state.is_arming())
        self.assertFalse(self.location_normal.arming_state.is_disarming())
        self.assertTrue(self.location_normal.arming_state.is_disarmed())
        self.assertFalse(self.location_normal.arming_state.is_armed_away())
        self.assertFalse(self.location_normal.arming_state.is_armed_custom_bypass())
        self.assertFalse(self.location_normal.arming_state.is_armed_home())
        self.assertFalse(self.location_normal.arming_state.is_armed_night())
        self.assertFalse(self.location_normal.arming_state.is_armed())
        self.assertFalse(self.location_normal.arming_state.is_pending())
        self.assertFalse(self.location_normal.arming_state.is_triggered_police())
        self.assertFalse(self.location_normal.arming_state.is_triggered_fire())
        self.assertFalse(self.location_normal.arming_state.is_triggered_gas())
        self.assertFalse(self.location_normal.arming_state.is_triggered())

        loc = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, self)
        response = deepcopy(RESPONSE_DISARMED)
        response["PanelMetadataAndStatus"] = METADATA_DISARMED_LOW_BATTERY
        loc._update_zones(response)
        assert loc.zones["1"].is_low_battery() is True

    def tests_update_status_none(self):
        """Test _update_status with None passed in."""
        loc = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, self)
        loc._update_status(deepcopy(RESPONSE_DISARMED))
        loc._update_partitions(deepcopy(RESPONSE_DISARMED))
        loc._update_zones(deepcopy(RESPONSE_DISARMED))

        self.assertTrue(loc.arming_state.is_disarmed())
        with pytest.raises(PartialResponseError):
            loc._update_status(None)

        data = deepcopy(RESPONSE_DISARMED)
        del data["PanelMetadataAndStatus"]["Partitions"]["PartitionInfo"]
        with pytest.raises(PartialResponseError):
            loc._update_partitions(data)
        del data["PanelMetadataAndStatus"]["Partitions"]
        with pytest.raises(PartialResponseError):
            loc._update_partitions(data)

        data = deepcopy(RESPONSE_DISARMED)
        del data["PanelMetadataAndStatus"]["Zones"]["ZoneInfo"]
        with pytest.raises(TotalConnectError):
            loc._update_zones(data)

        # See issue #112 when user's system returned zero zones,
        # and zeep set "Zones" to None
        data["PanelMetadataAndStatus"]["Zones"] = None
        with pytest.raises(TotalConnectError):
            loc._update_zones(data)

        del data["PanelMetadataAndStatus"]["Zones"]
        with pytest.raises(TotalConnectError):
            loc._update_zones(data)
        self.assertTrue(loc.arming_state.is_disarmed())

    def tests_set_zone_details(self):
        """Test set_zone_details with normal data passed in."""
        location = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, None)
        location._update_zone_details(RESPONSE_GET_ZONE_DETAILS_SUCCESS)
        assert len(location.zones) == 1

        location = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, None)
        # "Zones" is None, as seen in #112 and #205
        response = deepcopy(RESPONSE_GET_ZONE_DETAILS_SUCCESS)
        response["ZoneStatus"]["Zones"] = None
        location._update_zone_details(response)
        assert len(location.zones) == 0

        # "ZoneStatusInfoWithPartitionId" is None
        location = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, None)
        response = deepcopy(RESPONSE_GET_ZONE_DETAILS_SUCCESS)
        response["ZoneStatus"]["Zones"] = {"ZoneStatusInfoWithPartitionId": None}
        location._update_zone_details(response)
        assert len(location.zones) == 0

    def tests_auto_bypass_low_battery(self):
        """Test auto bypass of low battery zones."""

        mock_client = Mock()
        loc = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, mock_client)

        # should not try to bypass by default
        assert loc.auto_bypass_low_battery is False
        response = deepcopy(RESPONSE_DISARMED)
        response["PanelMetadataAndStatus"] = METADATA_DISARMED_LOW_BATTERY

        zbp = "total_connect_client.client.TotalConnectLocation.zone_bypass"
        with patch(zbp) as mock:
            loc._update_status(response)
            loc._update_partitions(response)
            loc._update_zones(response)
            assert mock.call_count == 0

        # now set to auto bypass
        loc.auto_bypass_low_battery = True

        # now update status with a low battery and ensure it is bypassed
        with patch(zbp) as mock:
            loc._update_status(response)
            loc._update_partitions(response)
            loc._update_zones(response)
            assert mock.call_count == 1

    def tests_set_usercode(self):
        """Test set_usercode."""

        mock_client = Mock()

        loc = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, mock_client)

        # should start with default usercode
        assert loc.usercode == DEFAULT_USERCODE

        # now set it with an invalid code
        mock_client.validate_usercode.return_value = False
        assert loc.set_usercode("0000") is False
        assert loc.usercode == DEFAULT_USERCODE
        assert mock_client.validate_usercode.call_count == 1

        # now set it with a valid code
        mock_client.validate_usercode.return_value = True
        assert loc.set_usercode("1234") is True
        assert loc.usercode == "1234"
        assert mock_client.validate_usercode.call_count == 2


def tests_update_status():
    """Test location._update_status()."""
    location = TotalConnectLocation(LOCATION_INFO_BASIC_NORMAL, Mock())

    # known arming state should not produce an error
    response = {
        "PanelMetadataAndStatus": METADATA_DISARMED,
        "ArmingState": ArmingState.DISARMED,
    }
    location._update_status(response)
    assert location.arming_state == ArmingState.DISARMED

    # unknown arming state should produce an error
    with pytest.raises(TotalConnectError):
        response["ArmingState"] = 99999
        location._update_status(response)



"""Test total_connect_client."""

import unittest
from unittest.mock import patch

import pytest
from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    # RESPONSE_ARMED_AWAY,
    # RESPONSE_ARMED_STAY,
    # RESPONSE_ARMED_STAY_NIGHT,
    RESPONSE_DISARMED,
    RESPONSE_FEATURE_NOT_SUPPORTED,
    RESPONSE_GET_ZONE_DETAILS_NONE,
    RESPONSE_GET_ZONE_DETAILS_SUCCESS,
    RESPONSE_UNKNOWN,
)

from total_connect_client.zone import ZoneStatus

from total_connect_client.exceptions import (
    BadResultCodeError,
    TotalConnectError,
)


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.client = create_client()
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]
        self.location = self.client.locations[self.location_id]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_zone_status(self):
        """Test zone_status."""
        responses = [RESPONSE_DISARMED]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start disarmed
            assert self.location.arming_state.is_disarmed()

            # ask for status of zone 1, which exists
            assert self.location.zone_status("1") == ZoneStatus.NORMAL

            # ask for status of zone 99, which does not exist
            with pytest.raises(TotalConnectError):
                self.location.zone_status("99")

    def tests_get_zone_details(self):
        """Test get_zone_details."""
        responses = [
            RESPONSE_GET_ZONE_DETAILS_SUCCESS,
            RESPONSE_FEATURE_NOT_SUPPORTED,
            RESPONSE_UNKNOWN,
            RESPONSE_GET_ZONE_DETAILS_NONE,
        ]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # first response is SUCCESS
            self.location.get_zone_details()
            # second response is FEATURE_NOT_SUPPORTED
            self.location.get_zone_details()
            # third response is UNKNOWN
            with pytest.raises(BadResultCodeError):
                self.location.get_zone_details()
            # third response is SUCCESS but with empty ZoneStatus
            # ...which we've seen before in the wild
            self.location.get_zone_details()



"""Test total_connect_client."""

from unittest.mock import patch

import pytest
from const import (
    RESPONSE_SESSION_DETAILS,
    RESPONSE_SESSION_DETAILS_EMPTY,
    RESPONSE_DISARMED,
    RESPONSE_GET_ZONE_DETAILS_SUCCESS,
    RESPONSE_PARTITION_DETAILS,
)

from tests.const import RESPONSE_PARTITION_DETAILS_TWO
from total_connect_client.client import TotalConnectClient
from total_connect_client.exceptions import TotalConnectError


def tests_init_usercodes_none():
    """Test init with usercodes == None."""
    responses = [
        RESPONSE_SESSION_DETAILS,
        RESPONSE_PARTITION_DETAILS,
        RESPONSE_GET_ZONE_DETAILS_SUCCESS,
        RESPONSE_DISARMED,
    ]

    with patch(
        "total_connect_client.client.TotalConnectClient.request", side_effect=responses
    ) as mock_request:
        mock_client = TotalConnectClient("username", "password", usercodes=None)
        assert mock_request.call_count == 1
        if mock_client.locations:  # force client to fetch them
            pass
        assert mock_request.call_count == 4

    assert not mock_client.usercodes


def tests_init_locations_empty():
    """Test init with no locations."""
    responses = [
        RESPONSE_SESSION_DETAILS_EMPTY,
    ]

    with patch(
        "total_connect_client.client.TotalConnectClient.request", side_effect=responses
    ) as mock_request, pytest.raises(TotalConnectError):

        mock_client = TotalConnectClient("username", "password", usercodes=None)
        assert mock_client.locations == {}
        assert mock_request.call_count == 1


def tests_init_usercodes_string():
    """Test init with usercodes == a string."""
    responses = [
        RESPONSE_SESSION_DETAILS,
        RESPONSE_PARTITION_DETAILS,
        RESPONSE_GET_ZONE_DETAILS_SUCCESS,
        RESPONSE_DISARMED,
    ]

    with patch(
        "total_connect_client.client.TotalConnectClient.request", side_effect=responses
    ):
        with pytest.raises(AttributeError):
            # string is not a valid type for usercodes (only dict)
            TotalConnectClient("username", "password", usercodes="123456")


def tests_init_partitions():
    """Test that partitions populate correctly."""
    # one partition
    responses = [
        RESPONSE_SESSION_DETAILS,
        RESPONSE_PARTITION_DETAILS,
        RESPONSE_GET_ZONE_DETAILS_SUCCESS,
        RESPONSE_DISARMED,
    ]

    with patch(
        "total_connect_client.client.TotalConnectClient.request", side_effect=responses
    ) as mock_request:
        mock_client = TotalConnectClient("username", "password", usercodes=None)
        assert mock_request.call_count == 1
        if mock_client.locations:  # force client to fetch them
            pass
        assert mock_request.call_count == 4

        assert len(mock_client.locations["123456"].partitions) == 1

    # two partitions
    responses = [
        RESPONSE_SESSION_DETAILS,
        RESPONSE_PARTITION_DETAILS_TWO,
        RESPONSE_GET_ZONE_DETAILS_SUCCESS,
        RESPONSE_DISARMED,
    ]

    with patch(
        "total_connect_client.client.TotalConnectClient.request", side_effect=responses
    ) as mock_request:
        mock_client = TotalConnectClient("username", "password", usercodes=None)
        assert mock_request.call_count == 1
        if mock_client.locations:  # force client to fetch them
            pass
        assert mock_request.call_count == 4

        assert len(mock_client.locations["123456"].partitions) == 2



"""Test client.zone_bypass()."""

import unittest
from unittest.mock import patch

import pytest
from common import create_client
from const import (
    LOCATION_INFO_BASIC_NORMAL,
    RESPONSE_DISARM_SUCCESS,
    TCC_REQUEST_METHOD,
)

from total_connect_client.const import _ResultCode
from total_connect_client.exceptions import FailedToBypassZone
from total_connect_client.zone import ZoneStatus

RESPONSE_ZONE_BYPASS_SUCCESS = {
    "ResultCode": _ResultCode.SUCCESS.value,
    "ResultData": "None",
}

# guessing on the response...don't know for sure
RESPONSE_ZONE_BYPASS_FAILURE = {
    "ResultCode": _ResultCode.FAILED_TO_BYPASS_ZONE.value,
    "ResultData": "None",
}


class TestTotalConnectClient(unittest.TestCase):
    """Test TotalConnectClient."""

    def setUp(self):
        """Test setup."""
        self.client = create_client()
        self.location_id = LOCATION_INFO_BASIC_NORMAL["LocationID"]

    def tearDown(self):
        """Test cleanup."""
        self.client = None

    def tests_zone_bypass_success(self):
        """Test Zone Bypass with a normal response."""
        location = self.client.locations[self.location_id]
        zone = location.zones["1"]
        responses = [RESPONSE_ZONE_BYPASS_SUCCESS]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start not bypassed
            assert zone.is_bypassed() is False

            # now bypass the zone
            location.zone_bypass("1")

            # should now be bypassed
            assert zone.is_bypassed() is True

    def tests_zone_bypass_failure(self):
        """Test Zone Bypass with a normal response."""
        location = self.client.locations[self.location_id]
        zone = location.zones["1"]
        responses = [RESPONSE_ZONE_BYPASS_FAILURE]
        with patch(
            "total_connect_client.client.TotalConnectClient.request",
            side_effect=responses,
        ):
            # should start not bypassed
            assert zone.is_bypassed() is False

            # try to bypass the zone
            with pytest.raises(FailedToBypassZone):
                location.zone_bypass("1")

            # should not be bypassed
            assert zone.is_bypassed() is False


def tests_bypass_all():
    """Test bypass_all."""
    client = create_client()
    location = client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]]
    zone = location.zones["1"]

    responses = [RESPONSE_DISARM_SUCCESS]
    with patch(TCC_REQUEST_METHOD, side_effect=responses) as mock_request:
        # no zones are faulted so should not bypass anything
        location.zone_bypass_all()
        assert mock_request.call_count == 0

        # mark the zone as faulted, now should bypass
        zone.status = ZoneStatus.FAULT
        location.zone_bypass_all()
        assert mock_request.call_count == 1


def tests_clear_bypass():
    """Test clear bypass, which is basically disarm()."""
    client = create_client()
    location = client.locations[LOCATION_INFO_BASIC_NORMAL["LocationID"]]

    responses = [RESPONSE_DISARM_SUCCESS]
    with patch(TCC_REQUEST_METHOD, side_effect=responses) as mock_request:
        # then clear bypass
        location.clear_bypass()

        assert mock_request.call_count == 1
